@@ -3,7 +3,7 @@
  *
  * T38Modem simulator - main program
  *
- * Copyright (c) 2001-2002 Vyacheslav Frolov
+ * Copyright (c) 2001-2005 Vyacheslav Frolov
  *
  * Open H323 Project
  *
@@ -24,6 +24,44 @@
  * Contributor(s): Vyacheslav Frolov
  *
  * $Log: main.cxx,v $
+ * Revision 1.40  2005/02/24 11:21:11  vfrolov
+ * Added ability to put list to --route option
+ *
+ * Revision 1.39  2005/02/04 10:18:48  vfrolov
+ * Fixed warnings for No Trace build
+ *
+ * Revision 1.38  2004/10/29 10:51:03  vfrolov
+ * Added missing OS class to trace
+ * Fixed MSVC compiler warning
+ *
+ * Revision 1.37  2004/10/27 13:22:42  vfrolov
+ * Added flags to call to PTrace::Initialise()
+ * Added version message to trace
+ *
+ * Revision 1.36  2004/10/20 13:40:03  vfrolov
+ * Put date and time to trace
+ *
+ * Revision 1.35  2004/07/07 12:38:32  vfrolov
+ * The code for pseudo-tty (pty) devices that communicates with fax application formed to PTY driver.
+ *
+ * Revision 1.34  2004/05/09 07:46:11  csoutheren
+ * Updated to compile with new PIsDescendant function
+ *
+ * Revision 1.33  2003/12/19 15:25:06  vfrolov
+ * Removed class AudioDelay (utilized PAdaptiveDelay)
+ * Renamed pmodemQ to pmodem_pool
+ * Fixed modem loss on no route
+ *
+ * Revision 1.32  2003/12/04 11:23:51  vfrolov
+ * Added "h323:" prefix to remoteParty
+ *
+ * Revision 1.31  2003/04/14 07:04:33  vfrolov
+ * Added missing --disable and --prefer options.
+ * Reported by Alexandre Aractingi.
+ *
+ * Revision 1.30  2003/04/07 06:52:43  vfrolov
+ * Fixed --save option
+ *
  * Revision 1.29  2002/12/20 10:12:46  vfrolov
  * Implemented tracing with PID of thread (for LinuxThreads)
  *   or ID of thread (for other POSIX Threads)
@@ -109,6 +147,7 @@
 #include "pmodem.h"
 #include "main.h"
 #include "g7231_fake.h"
+#include "drivers.h"
 
 PCREATE_PROCESS(T38Modem);
 
@@ -135,17 +174,27 @@ void T38Modem::Main()
        << " on " << GetOSClass() << ' ' << GetOSName()
        << " (" << GetOSVersion() << '-' << GetOSHardware() << ")\n\n";
 
+  if (!Initialise())
+    return;
+
+  for (;;)
+    PThread::Sleep(5000);
+}
+
+BOOL T38Modem::Initialise()
+{
   PConfigArgs args(GetArguments());
 
   args.Parse(
-	     "p-ptty:"
-	     "-route:"
+             "p-ptty:"
+             "-route:"
              "-redundancy:"
              "-old-asn."
 
              "F-fastenable."
              "T-h245tunneldisable."
-	     "G-g7231code."
+             "G-g7231code."
+             "D-disable:"            "P-prefer:"
 
              "g-gatekeeper:"         "n-no-gatekeeper."
              "-require-gatekeeper."  "-no-require-gatekeeper."
@@ -160,12 +209,8 @@ void T38Modem::Main()
              "t-trace."
              "o-output:"
 #endif
-             "r-run:"                "-no-run."
              "-save."
-#if PTRACING
-             "t-trace."
-#endif
-	     "u-username:"           "-no-username."
+             "u-username:"           "-no-username."
           , FALSE);
 
 #if PMEMORY_CHECK
@@ -175,7 +220,13 @@ void T38Modem::Main()
 
 #if PTRACING
   PTrace::Initialise(args.GetOptionCount('t'),
-                     args.HasOption('o') ? (const char *)args.GetOptionString('o') : NULL);
+                     args.HasOption('o') ? (const char *)args.GetOptionString('o') : NULL,
+                     PTrace::DateAndTime | PTrace::Thread | PTrace::Blocks);
+
+  PTRACE(1, GetName()
+      << " Version " << GetVersion(TRUE)
+      << " on " << GetOSClass() << " " << GetOSName()
+      << " (" << GetOSVersion() << '-' << GetOSHardware() << ")");
 #endif
 
   if (args.HasOption('h')) {
@@ -184,23 +235,22 @@ void T38Modem::Main()
         "Options:\n"
         "  -p --ptty [num@]tty[,...] : Pseudo ttys (mandatory).\n"
         "                              Can be used multiple times.\n"
-        "                              The tty should match to regexp\n"
-        "                                '" << PseudoModem::ttyPattern() << "'\n"
         "                              If tty prefixed by num@ then tty will\n"
         "                              accept incoming calls only\n"
         "                              for numbers with prefix num.\n"
         "                              Use none@tty to disable incoming calls.\n"
-        "  --route prefix@host       : Route numbers with prefix num to host.\n"
+        "                              See Drivers section for supported tty's formats.\n"
+        "  --route prefix@host[,...] : Route numbers with prefix num to host.\n"
         "                              Can be used multiple times.\n"
         "                              Discards prefix num from numbers.\n"
         "                              Use 'all' to route all numbers.\n"
-	"                              Mandatory if not using GK.\n"
+        "                              Mandatory if not using GK.\n"
         "  --redundancy I[L[H]]      : Set redundancy for error recovery for\n"
-	"                              (I)ndication, (L)ow speed and (H)igh\n"
+        "                              (I)ndication, (L)ow speed and (H)igh\n"
         "                              speed IFP packets.\n"
         "                              'I', 'L' and 'H' are digits.\n"
-        "  --old-asn                 : Use original ASN.1 sequence in T.38 (06/98) Annex A\n"
-        "                              (w/o CORRIGENDUM No. 1 fix).\n"
+        "  --old-asn                 : Use original ASN.1 sequence in T.38 (06/98)\n"
+        "                              Annex A (w/o CORRIGENDUM No. 1 fix).\n"
         "  -i --interface ip         : Bind to a specific interface.\n"
         "  --no-listenport           : Disable listen for incoming calls.\n"
         "  --listenport port         : Listen on a specific port.\n"
@@ -210,29 +260,44 @@ void T38Modem::Main()
         "  --require-gatekeeper      : Exit if gatekeeper discovery fails.\n"
         "  -F --fastenable           : Enable fast start.\n"
         "  -T --h245tunneldisable    : Disable H245 tunnelling.\n"
-	"  -G --g7231enable          : Enable G.723.1 codec, rather than G.711.\n"
-	"  -u --username str         : Set the local endpoint name to str.\n"
+        "  -G --g7231enable          : Enable G.723.1 codec, rather than G.711.\n"
+        "  -D --disable codec        : Disable the specified codec.\n"
+        "                              Can be used multiple times.\n"
+        "  -P --prefer codec         : Prefer the specified codec.\n"
+        "                              Can be used multiple times.\n"
+        "  -u --username str         : Set the local endpoint name to str.\n"
 #if PTRACING
         "  -t --trace                : Enable trace, use multiple times for more detail.\n"
         "  -o --output               : File for trace output, default is stderr.\n"
 #endif
-        "     --save                 : Save arguments in configuration file.\n"
-        "  -h --help                 : Display this help message.\n";
-    return;
+        "     --save                 : Save arguments in configuration file and exit.\n"
+        "  -h --help                 : Display this help message.\n"
+        "Drivers:\n";
+
+    PStringArray descriptions = PseudoModemDrivers::Descriptions();
+
+    for (PINDEX i = 0 ; i < descriptions.GetSize() ; i++) {
+      cout << 
+        "  " << descriptions[i] << "\n";
+    }
+    return FALSE;
   }
 
-  args.Save("save");
+  if (args.HasOption("save")) {
+    args.Save("save");
+    cout << "Arguments were saved in configuration file\n";
+    return FALSE;
+  }
 
-  MyH323EndPoint endpoint;
+  MyH323EndPoint *endpoint = new MyH323EndPoint();
 
   PString userName = "OpenH323 Answering Machine v" + GetVersion();
   if (args.HasOption('u'))
     userName = args.GetOptionString('u');
-  endpoint.SetLocalUserName(userName);
-
-  if (!endpoint.Initialise(args))
-    return;
+  endpoint->SetLocalUserName(userName);
 
+  if (!endpoint->Initialise(args))
+    return FALSE;
 
   // start the H.323 listener
   H323ListenerTCP * listener = NULL;
@@ -246,48 +311,47 @@ void T38Modem::Main()
     if (args.HasOption('i'))
       interfaceAddress = PIPSocket::Address(args.GetOptionString('i'));
 
-    listener  = new H323ListenerTCP(endpoint, interfaceAddress, listenPort);
+    listener = new H323ListenerTCP(*endpoint, interfaceAddress, listenPort);
 
-    if (!endpoint.StartListener(listener)) {
-      cout <<  "Could not open H.323 listener port on "
+    if (!endpoint->StartListener(listener)) {
+      cout << "Could not open H.323 listener port on "
            << listener->GetListenerPort() << endl;
       delete listener;
-      return;
+      return FALSE;
     }
   }
 
   if (args.HasOption('g')) {
     PString gkName = args.GetOptionString('g');
-    if (endpoint.SetGatekeeper(gkName, new H323TransportUDP(endpoint)))
-      cout << "Gatekeeper set: " << *endpoint.GetGatekeeper() << endl;
+    if (endpoint->SetGatekeeper(gkName, new H323TransportUDP(*endpoint)))
+      cout << "Gatekeeper set: " << *endpoint->GetGatekeeper() << endl;
     else {
       cout << "Error registering with gatekeeper at \"" << gkName << '"' << endl;
-      return;
+      return FALSE;
     }
   }
   else if (!args.HasOption('n')) {
     cout << "Searching for gatekeeper..." << flush;
-    if (endpoint.DiscoverGatekeeper(new H323TransportUDP(endpoint)))
-      cout << "\nGatekeeper found: " << *endpoint.GetGatekeeper() << endl;
+    if (endpoint->DiscoverGatekeeper(new H323TransportUDP(*endpoint)))
+      cout << "\nGatekeeper found: " << *endpoint->GetGatekeeper() << endl;
     else {
       cout << "\nNo gatekeeper found." << endl;
       if (args.HasOption("require-gatekeeper"))
-        return;
+        return FALSE;
     }
   }
 
-  cout << "Waiting for incoming calls for \"" << endpoint.GetLocalUserName() << '"'
+  cout << "Waiting for incoming calls for \"" << endpoint->GetLocalUserName() << '"'
        << PString(listener ? "" : " disabled") << endl;
 
-  for (;;)
-    PThread::Sleep(5000);
+  return TRUE;
 }
 
 ///////////////////////////////////////////////////////////////
 
 MyH323EndPoint::MyH323EndPoint()
 {
-  pmodemQ = new PseudoModemQ();
+  pmodem_pool = new PseudoModemQ();
   //autoStartTransmitFax = TRUE;
 
   in_redundancy = -1;
@@ -296,9 +360,9 @@ MyH323EndPoint::MyH323EndPoint()
   old_asn = FALSE;
 }
 
-void MyH323EndPoint::OnMyCallback(PObject &from, INT extra)
+void MyH323EndPoint::OnMyCallback(PObject &from, INT myPTRACE_PARAM(extra))
 {
-  if (from.IsDescendant(PStringToString::Class()) ) {
+  if (PIsDescendant(&from, PStringToString) ) {
     PStringToString &request = (PStringToString &)from;
     PString command = request("command");
 
@@ -307,15 +371,15 @@ void MyH323EndPoint::OnMyCallback(PObject &from, INT extra)
     PString modemToken = request("modemtoken");
     PString response = "reject";
   
-    if (command == "dial" ) {
-      PseudoModem *modem = pmodemQ->Dequeue(modemToken);
-      if (modem != NULL ) {
+    if (command == "dial") {
+      PseudoModem *modem = pmodem_pool->Dequeue(modemToken);
+      if (modem != NULL) {
         PString num = request("number");
         PString remote;
 
-	if (GetGatekeeper() != NULL) {
+        if (GetGatekeeper() != NULL) {
           remote = num;
-	} else {
+        } else {
           for( PINDEX i = 0 ; i < routes.GetSize() ; i++ ) {
             PString r = routes[i];
             PStringArray rs = r.Tokenise("@", FALSE);
@@ -335,8 +399,8 @@ void MyH323EndPoint::OnMyCallback(PObject &from, INT extra)
           if (!remote.IsEmpty()) {
             num += "@" + remote;
             if ((num.Find(':') == P_MAX_INDEX) && (connectPort != H323EndPoint::DefaultTcpPort))
-	      num += psprintf(":%i", connectPort);
-	  }
+              num += psprintf(":%i", connectPort);
+          }
         }
 
         if (remote.IsEmpty()) 
@@ -347,29 +411,29 @@ void MyH323EndPoint::OnMyCallback(PObject &from, INT extra)
           PString LocalPartyName = request("localpartyname");
           PString callToken;
 
-          MakeCall(num, callToken, (void *)(LocalPartyName.IsEmpty() ? NULL : (const char*)LocalPartyName));
+          MakeCall("h323:" + num, callToken, (void *)(LocalPartyName.IsEmpty() ? NULL : (const char*)LocalPartyName));
 
           request.SetAt("calltoken", callToken);
           H323Connection * _conn = FindConnectionWithLock(callToken);
 
-          if (_conn == NULL ) 
-            pmodemQ->Enqueue(modem);
-          else {
+          if (_conn != NULL) {
             cout << "O/G connection to " << num;
             if (!LocalPartyName.IsEmpty())
               cout << " from " << LocalPartyName;
             cout << endl;
-            PAssert(_conn->IsDescendant(MyH323Connection::Class()), PInvalidCast);
+            PAssert(PIsDescendant(_conn, MyH323Connection), PInvalidCast);
             MyH323Connection *conn = (MyH323Connection *)_conn;
-            if( conn->Attach(modem) )
+            if (conn->Attach(modem)) {
               response = "confirm";
-            else
-              pmodemQ->Enqueue(modem);
+              modem = NULL;
+            }
             _conn->Unlock();
           }
         }
+        if (modem != NULL)
+          pmodem_pool->Enqueue(modem);
       }
-    } else if( command == "answer" ) {
+    } else if (command == "answer") {
       PString callToken = request("calltoken");
       H323Connection * _conn = FindConnectionWithLock(callToken);
       if( _conn != NULL ) {
@@ -377,7 +441,7 @@ void MyH323EndPoint::OnMyCallback(PObject &from, INT extra)
         _conn->Unlock();
         response = "confirm";
       }
-    } else if( command == "requestmode" ) {
+    } else if (command == "requestmode") {
       PString callToken = request("calltoken");
       H323Connection * _conn = FindConnectionWithLock(callToken);
       if( _conn != NULL ) {
@@ -393,14 +457,19 @@ void MyH323EndPoint::OnMyCallback(PObject &from, INT extra)
         }
         _conn->Unlock();
       }
-    } else if( command == "clearcall" ) {
+    } else if (command == "clearcall") {
       PString callToken = request("calltoken");
       if( ClearCall(callToken) ) {
         response = "confirm";
       }
+    } else if (command == "addmodem") {
+      if (pmodem_pool->Enqueue(modemToken)) {
+        response = "confirm";
+      }
     }
+
     request.SetAt("response", response);
-    
+
     myPTRACE(1, "MyH323EndPoint::OnMyCallback request={\n" << request << "}");
   } else {
     myPTRACE(1, "MyH323EndPoint::OnMyCallback unknown class " << from.GetClass() << " extra=" << extra);
@@ -417,13 +486,13 @@ H323Connection * MyH323EndPoint::CreateConnection(unsigned callReference, void *
 
 PseudoModem * MyH323EndPoint::PMAlloc(const PString &number) const
 {
-  return pmodemQ->DequeueWithRoute(number);
+  return pmodem_pool->DequeueWithRoute(number);
 }
 
 void MyH323EndPoint::PMFree(PseudoModem *pmodem) const
 {
-  if( pmodem != NULL )
-    pmodemQ->Enqueue(pmodem);
+  if (pmodem != NULL)
+    pmodem_pool->Enqueue(pmodem);
 }
 
 void MyH323EndPoint::SetOptions(MyH323Connection &/*conn*/, OpalT38Protocol *t38handler) const
@@ -431,7 +500,7 @@ void MyH323EndPoint::SetOptions(MyH323Connection &/*conn*/, OpalT38Protocol *t38
   // TODO: make it per host
 
   if (t38handler != NULL) {
-    PAssert(t38handler->IsDescendant(T38Engine::Class()), PInvalidCast);
+    PAssert(PIsDescendant(t38handler, T38Engine), PInvalidCast);
     ((T38Engine *)t38handler)->SetRedundancy(in_redundancy, ls_redundancy, hs_redundancy);
     if (old_asn)
       ((T38Engine *)t38handler)->SetOldASN();
@@ -451,7 +520,7 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
 
   if (args.HasOption("route")) {
     PString r = args.GetOptionString("route");
-    routes = r.Tokenise("\r\n", FALSE);
+    routes = r.Tokenise(",\r\n", FALSE);
 
     cout << "Route O/G calls:\n";
 
@@ -482,7 +551,7 @@ BOOL MyH323EndPoint::Initialise(PConfigArgs & args)
         tty = atty[0];
       }
 
-      if (!pmodemQ->CreateModem(tty, r, PCREATE_NOTIFIER(OnMyCallback)))
+      if (!pmodem_pool->CreateModem(tty, r, PCREATE_NOTIFIER(OnMyCallback)))
         cout << "Can't create modem for " << tty << endl;
     }
   }
@@ -536,7 +605,7 @@ MyH323Connection::~MyH323Connection()
 
   if (pmodem != NULL) {
     if (t38handler != NULL) {
-      PAssert(t38handler->IsDescendant(T38Engine::Class()), PInvalidCast);
+      PAssert(PIsDescendant(t38handler, T38Engine), PInvalidCast);
       pmodem->Detach((T38Engine *)t38handler);
     }
 
@@ -599,7 +668,7 @@ OpalT38Protocol * MyH323Connection::CreateT38ProtocolHandler()
 }
 
 H323Connection::AnswerCallResponse
-     MyH323Connection::OnAnswerCall(const PString & caller,
+     MyH323Connection::OnAnswerCall(const PString & /*caller*/,
                                     const H323SignalPDU & setupPDU,
                                     H323SignalPDU & /*connectPDU*/)
 {
@@ -625,13 +694,13 @@ H323Connection::AnswerCallResponse
     myPTRACE(1, "... denied (all modems busy)");
     return AnswerCallDenied;
   }
-  
+
   if (!Attach(_pmodem)) {
     myPTRACE(1, "... denied (internal error)");
-    ep.PMFree(pmodem);
+    ep.PMFree(_pmodem);
     return AnswerCallDenied;
   }
-  
+
   PString old = PThread::Current()->GetThreadName();
   RenameCurrentThread(pmodem->ptyName() + "(c)");
   PTRACE(2, "MyH323Connection::AnswerCallResponse old ThreadName=" << old);
@@ -697,10 +766,7 @@ void MyH323Connection::OnClosedLogicalChannel(const H323Channel & channel)
 BOOL MyH323Connection::OpenAudioChannel(BOOL isEncoding, unsigned /* bufferSize */, H323AudioCodec & codec)
 {
   //codec.SetSilenceDetectionMode(H323AudioCodec::NoSilenceDetection);
-  
-  PStringStream codecName;
-  codecName << codec;
-  
+
   PTRACE(2, "MyH323Connection::OpenAudioChannel " << codec);
 
   PWaitAndSignal mutex(connMutex);
@@ -778,44 +844,5 @@ BOOL AudioWrite::Close()
   closed = TRUE;
   return TRUE;
 }
-
-///////////////////////////////////////////////////////////////
-
-AudioDelay::AudioDelay()
-{
-  firstTime = TRUE;
-  error = 0;
-}
-
-void AudioDelay::Restart()
-{
-  firstTime = TRUE;
-}
-
-BOOL AudioDelay::Delay(int frameTime)
-{
-  if (firstTime) {
-    firstTime = FALSE;
-    previousTime = PTime();
-    return TRUE;
-  }
-
-  error += frameTime;
-
-  PTime now;
-  PTimeInterval delay = now - previousTime;
-  error -= (int)delay.GetMilliSeconds();
-  previousTime = now;
-
-  if (error > 0)
-#ifdef P_LINUX
-    usleep(error * 1000);
-#else
-    PThread::Sleep(error);
-#endif
-
-  return error <= -frameTime;
-}
-
 // End of File ///////////////////////////////////////////////////////////////
 
