@@ -3,7 +3,7 @@
  *
  * T38FAX Pseudo Modem
  *
- * Copyright (c) 2001-2002 Vyacheslav Frolov
+ * Copyright (c) 2001-2005 Vyacheslav Frolov
  *
  * Open H323 Project
  *
@@ -24,6 +24,40 @@
  * Contributor(s): Equivalence Pty ltd
  *
  * $Log: pmodeme.cxx,v $
+ * Revision 1.28  2005/02/16 12:14:47  vfrolov
+ * Send CONNECT just before data for AT+FRM command
+ *
+ * Revision 1.27  2005/02/10 10:35:18  vfrolov
+ * Fixed AT prefix searching
+ *
+ * Revision 1.26  2005/02/03 11:32:11  vfrolov
+ * Fixed MSVC compile warnings
+ *
+ * Revision 1.25  2005/02/01 11:43:46  vfrolov
+ * Implemented ATV0 command (numeric format for result codes)
+ * Implemented AT+FMI?, AT+FMM? and AT+FMR? commands
+ * Added stubs for ATBn, ATX3 and AT+FCLASS=0 commands
+ * Added stub for AT+FLO command
+ *
+ * Revision 1.24  2004/10/27 13:36:26  vfrolov
+ * Decreased binary, DLE, and callback tracing
+ *
+ * Revision 1.23  2004/07/06 16:07:24  vfrolov
+ * Included ptlib.h for precompiling
+ *
+ * Revision 1.22  2004/06/24 17:20:22  vfrolov
+ * Added stub for ATXn command
+ *
+ * Revision 1.21  2004/05/09 07:46:11  csoutheren
+ * Updated to compile with new PIsDescendant function
+ *
+ * Revision 1.20  2004/03/01 17:14:34  vfrolov
+ * Fixed binary log in command mode
+ *
+ * Revision 1.19  2003/12/04 16:09:51  vfrolov
+ * Implemented FCS generation
+ * Implemented ECM support
+ *
  * Revision 1.18  2003/01/08 16:58:58  vfrolov
  * Added cbpOutBufNoFull and isOutBufFull()
  *
@@ -94,9 +128,11 @@
  *
  */
 
+#include <ptlib.h>
 #include "pmodemi.h"
 #include "pmodeme.h"
 #include "dle.h"
+#include "fcs.h"
 #include "t38engine.h"
 #include "version.h"
 
@@ -117,6 +153,7 @@ static const char Revision[] = TOSTR(MAJOR_VERSION) "." TOSTR(MINOR_VERSION) "."
 class Profile
 {
   enum {
+    FloReg = 49,
     CidModeReg = 50,
     MaxReg = 50,
     MaxBit = 7
@@ -154,7 +191,7 @@ class Profile
       GetBit(23, 7, val);
       return val;
     }
-    
+
     void CidMode(BYTE val) {
       SetReg(CidModeReg, val);
     }
@@ -165,6 +202,16 @@ class Profile
       return val;
     }
 
+    void Flo(BYTE val) {
+      SetReg(FloReg, val);
+    }
+
+    BYTE Flo() const {
+      BYTE val;
+      GetReg(FloReg, val);
+      return val;
+    }
+
     BOOL SetBit(PINDEX r, PINDEX b, BOOL val) {
       if( !ChkRB(r, b) ) return FALSE;
       BYTE msk = MaskB(b);
@@ -173,7 +220,10 @@ class Profile
       return TRUE;
     }
     BOOL GetBit(PINDEX r, PINDEX b, BOOL &val) const {
-      if( !ChkRB(r, b) ) return FALSE;
+      if (!ChkRB(r, b)) {
+        val = 0;
+        return FALSE;
+      }
       BYTE msk = MaskB(b);
       val = (S[r] & msk) ? TRUE : FALSE;
       return TRUE;
@@ -184,7 +234,10 @@ class Profile
       return TRUE;
     }
     BOOL GetReg(PINDEX r, BYTE &val) const {
-      if( !ChkR(r) ) return FALSE;
+      if (!ChkR(r)) {
+        val = 0;
+        return FALSE;
+      }
       val = S[r];
       return TRUE;
     }
@@ -196,9 +249,12 @@ class Profile
       return TRUE;
     }
     BOOL GetBits(PINDEX r, PINDEX bl, PINDEX bh, BYTE &val) const {
-      if( !ChkRBB(r, bl, bh) ) return FALSE;
+      if (!ChkRBB(r, bl, bh)) {
+        val = 0;
+        return FALSE;
+      }
       BYTE msk = MaskBB(bl, bh);
-      val = (S[r] & msk) >> bl;
+      val = BYTE((S[r] & msk) >> bl);
       return TRUE;
     }
     
@@ -221,7 +277,7 @@ class Profile
       return "\x01\x02\x04\x08\x10\x20\x40\x80"[b];
     }
     static BYTE MaskBB(PINDEX bl, PINDEX bh) { // bl <= bh
-      return ("\x01\x03\x07\x0F\x1F\x3F\x7F\xFF"[bh - bl]) << bl;
+      return BYTE(("\x01\x03\x07\x0F\x1F\x3F\x7F\xFF"[bh - bl]) << bl);
     }
     
     BYTE S[MaxReg + 1];	// S-registers
@@ -276,6 +332,9 @@ class Timeout : public PTimer
     PMutex Mutex;
 };
 ///////////////////////////////////////////////////////////////
+#define DeclareResultCode(name, v0, v1)	\
+  PString name() const { return P.asciiResultCodes() ? (v1) : (v0); }
+///////////////////////////////////////////////////////////////
 class ModemEngineBody : public PObject
 {
     PCLASSINFO(ModemEngineBody, PObject);
@@ -330,10 +389,10 @@ class ModemEngineBody : public PObject
       return t38engine && t38engine->isOutBufFull();
     }
   //@}
-  
+
   protected:
     BOOL Echo() const { return P.Echo(); }
-    BOOL HandleClass1Cmd(const char **ppCmd, PString &resp);
+    BOOL HandleClass1Cmd(const char **ppCmd, PString &resp, BOOL &ok, BOOL &crlf);
     void HandleCmd(const PString &cmd, PString &resp);
 
     void ResetDleData() {
@@ -349,7 +408,7 @@ class ModemEngineBody : public PObject
       ResetDleData();
       state = stSend;
       if( t38engine && t38engine->SendStart(dataType, br) ) {
-        resp += "\r\nCONNECT\r\n";
+        resp = RC_CONNECT();
         return TRUE;
       } else {
         state = stCommand;
@@ -398,6 +457,7 @@ class ModemEngineBody : public PObject
     DLEData dleData;
     PINDEX dataCount;
     BOOL moreFrames;
+    FCS fcs;
 
     Profile P;
     Profile Profiles[1];
@@ -408,6 +468,19 @@ class ModemEngineBody : public PObject
     DeclareStringParam(SrcNum)
     DeclareStringParam(DstNum)
 
+    DeclareResultCode(RC_PREF,            "", "\r\n")
+
+    DeclareResultCode(RC_OK,           "0\r", "OK\r\n")
+    DeclareResultCode(RC_CONNECT,      "1\r", "CONNECT\r\n")
+    DeclareResultCode(RC_RING,         "2\r", "RING\r\n")
+    DeclareResultCode(RC_NO_CARRIER,   "3\r", "NO CARRIER\r\n")
+    DeclareResultCode(RC_ERROR,        "4\r", "ERROR\r\n")
+    DeclareResultCode(RC_CONNECT_1200, "5\r", "CONNECT 1200\r\n")
+    DeclareResultCode(RC_NO_DIALTONE,  "6\r", "NO DIALTONE\r\n")
+    DeclareResultCode(RC_BUSY,         "7\r", "BUSY\r\n")
+    DeclareResultCode(RC_NO_ANSWER,    "8\r", "NO ANSWER\r\n")
+    DeclareResultCode(RC_RINGING,      "9\r", "RINGING\r\n")
+    DeclareResultCode(RC_FCERROR,    "+F4\r", "+FCERROR\r\n")
 };
 ///////////////////////////////////////////////////////////////
 
@@ -660,8 +733,8 @@ void ModemEngineBody::Detach(T38Engine *_t38engine)
 
 void ModemEngineBody::OnMyCallback(PObject &from, INT extra)
 {
-  PTRACE(1, "ModemEngineBody::OnMyCallback " << from.GetClass() << " " << extra);
-  if( from.IsDescendant(T38Engine::Class()) ) {
+  PTRACE(extra < 0 ? 1 : 3, "ModemEngineBody::OnMyCallback " << from.GetClass() << " " << extra);
+  if (PIsDescendant(&from, T38Engine) ) {
     PWaitAndSignal mutexWait(Mutex);
     if( extra == seq ) {
       switch( state ) {
@@ -718,7 +791,7 @@ static int ParseNum(const char **ppCmd,
     return num;
 }
 
-BOOL ModemEngineBody::HandleClass1Cmd(const char **ppCmd, PString &resp)
+BOOL ModemEngineBody::HandleClass1Cmd(const char **ppCmd, PString &resp, BOOL &ok, BOOL &crlf)
 {
   BOOL T;
   
@@ -755,12 +828,15 @@ BOOL ModemEngineBody::HandleClass1Cmd(const char **ppCmd, PString &resp)
         switch( **ppCmd ) {
           case '?':
             (*ppCmd)++;
-              resp += "\r\n0-255\r\nOK\r\n";
+            resp += "\r\n0-255";
+            crlf = TRUE;
             break;
           default:
             {
               int dms = ParseNum(ppCmd);
               if( dms >= 0 ) {
+                ok = FALSE;
+
                 PWaitAndSignal mutexWait(Mutex);
                 dataType = dt;
                 state = stSend;
@@ -790,18 +866,18 @@ BOOL ModemEngineBody::HandleClass1Cmd(const char **ppCmd, PString &resp)
           case '?':
             (*ppCmd)++;
             if( dt == T38Engine::dtRaw )
-              resp += "\r\n24,48,72,73,74,96,97,98,121,122,145,146\r\nOK\r\n";
+              resp += "\r\n24,48,72,73,74,96,97,98,121,122,145,146";
             else
-              resp += "\r\n3\r\nOK\r\n";
+              resp += "\r\n3";
+            crlf = TRUE;
             break;
           default:
             {
               int br = ParseNum(ppCmd);
-              
-              if( (dt == T38Engine::dtRaw && br == 3) || (dt == T38Engine::dtHdlc && br != 3) ) {
+
+              if ((dt == T38Engine::dtRaw && br == 3) || (dt == T38Engine::dtHdlc && br != 3))
                 return FALSE;
-	      }
-                
+
               switch( br ) {
                 case 3:
                 case 24:
@@ -816,10 +892,26 @@ BOOL ModemEngineBody::HandleClass1Cmd(const char **ppCmd, PString &resp)
                 case 122:
                 case 145:
                 case 146:
-                  if( T )
-                    return SendStart(dt, br, resp);
-                  else
-                    return RecvStart(dt, br);
+                  ok = FALSE;
+                  {
+                    PString _resp;
+                    BOOL res = T ? SendStart(dt, br, _resp) : RecvStart(dt, br);
+
+                    if (_resp.GetLength()) {
+                      if (crlf)
+                        resp += "\r\n";
+                      else
+                        resp += RC_PREF();
+
+                      crlf = FALSE;
+
+                      resp += _resp;
+                    }
+
+                    if (!res)
+                      PThread::Sleep(100);	// workaround
+                    return res;
+                  }
                 default:
                   return FALSE;
               }
@@ -847,20 +939,47 @@ BOOL ModemEngineBody::HandleClass1Cmd(const char **ppCmd, PString &resp)
 
 void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
 {
-  if (!(cmd.Left(2) *= "AT"))
-    return;
-    
-  PString ucmd = cmd.ToUpper();
-  const char * pCmd = ucmd;
-  pCmd += 2;
+
+  PINDEX i;
+
+  for (i = 0 ;  ; i++) {
+    i = cmd.FindOneOf("Aa", i);
+
+    if (i == P_MAX_INDEX) {
+      myPTRACE(1, "--> " << cmd.GetLength() << " bytes of binary");
+      return;
+    }
+
+    PString at = cmd.Mid(i, 2);
+
+    if (at == "AT" || at == "at")
+      break;
+  }
+
+  if (i) {
+    PBYTEArray bin((const BYTE *)(const char *)cmd, i);
+
+    myPTRACE(1, "--> " << PRTHEX(bin));
+  }
+
+  const char *pCmd;
+
+  pCmd = ((const char *)cmd) + i;
+
+  myPTRACE(1, "--> " << pCmd);
+
+  pCmd += 2;  // skip AT
+
+  PString ucmd = PString(pCmd).ToUpper();
+  pCmd = ucmd;
 
   BOOL err = FALSE;
   BOOL ok = TRUE;
-  BOOL crlf = TRUE;
-  
-  while (state == stCommand && !err && *pCmd)  {
+  BOOL crlf = FALSE;
+
+  while (state == stCommand && !err && *pCmd) {
       switch( *pCmd++ ) {
-	case ' ':
+        case ' ':
           break;
         case 'A':	// Accept incoming call
           ok = FALSE;
@@ -900,10 +1019,15 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                 state = stConnectHandle;
                 parent.SignalDataReady();
               }
-              crlf = FALSE;
             }
           }
           break;
+        case 'B':       // Turn ITU-T V.22/BELL 212A
+          if (ParseNum(&pCmd, 0, 1) >= 0) {
+          } else {
+            err = TRUE;
+          }
+          break;
         case 'D':	// Dial
           ok = FALSE;
           forceFaxMode = FALSE;
@@ -966,7 +1090,6 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
               
               if (response == "confirm") {
                 CallToken(request("calltoken"));
-                crlf = FALSE;
               } else {
                 callDirection = cdUndefined;
                 forceFaxMode = FALSE;
@@ -975,9 +1098,9 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                 if (response == "reject") {
                   PString diag = request("diag");
                   if( diag == "noroute" )
-                    resp += "\r\nBUSY";
+                    resp += RC_BUSY();
                   else
-                    resp += "\r\nNO DIALTONE";
+                    resp += RC_NO_DIALTONE();
                 } else {
                   err = TRUE;
                 }
@@ -1002,18 +1125,20 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
             
             switch( val ) {
               case 0:
-                resp += "\r\n";
-                resp += Model;
+                resp += "\r\n" + PString(Model);
+                crlf = TRUE;
                 break;
               case 3:
-                resp += "\r\n";
-                resp += Manufacturer;
+                resp += "\r\n" + PString(Manufacturer);
+                crlf = TRUE;
                 break;
               case 8:
                 resp += "\r\nNMBR = " + SrcNum();
+                crlf = TRUE;
                 break;
               case 9:
                 resp += "\r\nNDID = " + DstNum();
+                crlf = TRUE;
                 break;
               default:
                 if( val < 0 )
@@ -1065,6 +1190,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                 
                   if( P.GetReg(r, val) ) {
                     resp.sprintf("\r\n%3.3u", (unsigned)val);
+                    crlf = TRUE;
                   } else {
                     err = TRUE;
                   }
@@ -1090,6 +1216,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                 
                           if( P.GetBit(r, b, val) ) {
                             resp.sprintf("\r\n%u", (unsigned)val);
+                            crlf = TRUE;
                           } else {
                             err = TRUE;
                           }
@@ -1114,6 +1241,12 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
         case 'V':	// Numeric or ASCII result codes
           ToSBit(asciiResultCodes);
           break;
+        case 'X':	// Which result codes
+          if (ParseNum(&pCmd, 0, 1) >= 3) {
+          } else {
+            err = TRUE;
+          }
+          break;
         case 'Z':	// Load Registers from Profile
           {
             int val = ParseNum(&pCmd, 0, 1, sizeof(Profiles)/sizeof(Profiles[0]) - 1);
@@ -1139,6 +1272,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                           case '?':
                             pCmd++;
                             resp += "\r\n0";
+                            crlf = TRUE;
                             break;
                           default:
                             switch( ParseNum(&pCmd) ) {
@@ -1151,6 +1285,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                         break;
                       case '?':
                         resp += "\r\n0";
+                        crlf = TRUE;
                         break;
                       default:
                         err = TRUE;
@@ -1168,9 +1303,13 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                           case '?':
                             pCmd++;
                             resp += "\r\n1";	// "\r\n1,8"
+                            crlf = TRUE;
                             break;
                           default:
                             switch( ParseNum(&pCmd) ) {
+                              case 0:
+                                P.ModemClass("0");
+                                break;
                               case 1:
                                 P.ModemClass("1");
                                 break;
@@ -1184,6 +1323,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                         break;
                       case '?':
                         resp += "\r\n" + P.ModemClass();
+                        crlf = TRUE;
                         break;
                       default:
                         err = TRUE;
@@ -1192,29 +1332,110 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                     err = TRUE;
                   }
                   break;
+                case 'L':
+                  switch (*pCmd++) {
+                    case 'O': // +FLO
+                      switch (*pCmd++) {
+                        case '=':
+                          switch (*pCmd) {
+                            case '?':
+                              pCmd++;
+                              resp += "\r\n0-2";
+                              crlf = TRUE;
+                              break;
+                            default:
+                              {
+                                int val = ParseNum(&pCmd);
+                                switch (val) {
+                                  case 0:
+                                  case 1:
+                                  case 2:
+                                    P.Flo((BYTE)val);
+                                    break;
+                                  default:
+                                    err = TRUE;
+                                }
+                              }
+                          }
+                          break;
+                        case '?':
+                          resp.sprintf("\r\n%u", (unsigned)P.Flo());
+                          crlf = TRUE;
+                          break;
+                        default:
+                          err = TRUE;
+                      }
+                      break;
+                    default:
+                      err = TRUE;
+                  }
+                  break;
                 case 'M':
-                  if( strncmp(pCmd, "FR", 2) == 0 ) {	// +FMFR
-                    pCmd += 2;
-                    switch( *pCmd++ ) {
-                      case '?':
-                        resp += "\r\n";
-                        resp += Manufacturer;
-                        break;
+                  switch (*pCmd++) {
+                    case 'D':
+                      switch (*pCmd++) {
+                        case 'L': // +FMDL
+                          switch (*pCmd++) {
+                            case '?':
+                              resp += "\r\n" + PString(Model);
+                              crlf = TRUE;
+                              break;
+                            default:
+                              err = TRUE;
+                          }
+                          break;
                       default:
-                        err = TRUE;
-                    }
-                  } else if( strncmp(pCmd, "DL", 2) == 0 ) {	// +FMDL
-                    pCmd += 2;
-                    switch( *pCmd++ ) {
-                      case '?':
-                        resp += "\r\n";
-                        resp += Model;
-                        break;
+                          err = TRUE;
+                      }
+                      break;
+                    case 'F':
+                      switch (*pCmd++) {
+                        case 'R': // +FMFR
+                          switch (*pCmd++) {
+                            case '?':
+                              resp += "\r\n" + PString(Manufacturer);
+                              crlf = TRUE;
+                              break;
+                            default:
+                              err = TRUE;
+                          }
+                          break;
                       default:
-                        err = TRUE;
-                    }
-                  } else {
-                    err = TRUE;
+                          err = TRUE;
+                      }
+                      break;
+                    case 'I': // +FMI
+                      switch (*pCmd++) {
+                        case '?':
+                          resp += "\r\n" + PString(Manufacturer);
+                          crlf = TRUE;
+                          break;
+                        default:
+                          err = TRUE;
+                      }
+                      break;
+                    case 'M': // +FMM
+                      switch (*pCmd++) {
+                        case '?':
+                          resp += "\r\n" + PString(Model);
+                          crlf = TRUE;
+                          break;
+                        default:
+                          err = TRUE;
+                      }
+                      break;
+                    case 'R': // +FMR
+                      switch (*pCmd++) {
+                        case '?':
+                          resp += "\r\n" + PString(Revision);
+                          crlf = TRUE;
+                          break;
+                        default:
+                          err = TRUE;
+                      }
+                      break;
+                    default:
+                      err = TRUE;
                   }
                   break;
                 case 'R':
@@ -1223,10 +1444,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                     case 'H':				// +FRH
                     case 'S':				// +FRS
                       pCmd++;
-                      ok = FALSE;
-                      if( HandleClass1Cmd(&pCmd, resp) )
-                        crlf = FALSE;
-                      else
+                      if (!HandleClass1Cmd(&pCmd, resp, ok, crlf))
                         err = TRUE;
                       break;
                     default:
@@ -1234,8 +1452,8 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                         pCmd += 2;
                         switch( *pCmd++ ) {
                           case '?':
-                            resp += "\r\n";
-                            resp += Revision;
+                            resp += "\r\n" + PString(Revision);
+                            crlf = TRUE;
                             break;
                           default:
                             err = TRUE;
@@ -1251,10 +1469,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                     case 'H':				// +FTH
                     case 'S':				// +FTS
                       pCmd++;
-                      ok = FALSE;
-                      if( HandleClass1Cmd(&pCmd, resp) )
-                        crlf = FALSE;
-                      else
+                      if (!HandleClass1Cmd(&pCmd, resp, ok, crlf))
                         err = TRUE;
                       break;
                     default:
@@ -1311,7 +1526,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
           }
           break;
         case '#':
-          if( strncmp(pCmd, "CID", 3) == 0 ) {		// #CID
+          if (strncmp(pCmd, "CID", 3) == 0) {		// #CID
             pCmd += 3;
             switch( *pCmd++ ) {
               case '=':
@@ -1319,14 +1534,15 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                   case '?':
                     pCmd++;
                     resp += "\r\n(0,10)";
+                    crlf = TRUE;
                     break;
                   default:
                     {
                       int val = ParseNum(&pCmd);
-                      switch( val ) {
+                      switch (val) {
                         case 0:
                         case 10:
-                          P.CidMode(val);
+                          P.CidMode((BYTE)val);
                           break;
                         default:
                           err = TRUE;
@@ -1336,6 +1552,7 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
                 break;
               case '?':
                 resp.sprintf("\r\n%u", (unsigned)P.CidMode());
+                crlf = TRUE;
                 break;
               default:
                 err = TRUE;
@@ -1348,13 +1565,21 @@ void ModemEngineBody::HandleCmd(const PString & cmd, PString & resp)
           err = TRUE;
       }
   }
-  if( err ) {
-    resp += "\r\nERROR";
-  } else if( ok ) {
-    resp += "\r\nOK";
-  }
-  if( crlf )
+
+  if (crlf)
     resp += "\r\n";
+
+  if (err) {
+    if (!crlf)
+      resp += RC_PREF();
+    resp += RC_ERROR();
+  }
+  else
+  if (ok) {
+    if (!crlf)
+      resp += RC_PREF();
+    resp += RC_OK();
+  }
 }
 
 void ModemEngineBody::HandleData(const PBYTEArray &buf, PBYTEArray &bresp)
@@ -1390,22 +1615,21 @@ void ModemEngineBody::HandleData(const PBYTEArray &buf, PBYTEArray &bresp)
               }
               len--;
               pBuf++;
-    
-              if( Echo() )
+
+              if (Echo())
                 bresp.Concatenate(PBYTEArray((const BYTE *)"\r", 1));
-        
+
               PString resp;
-        
-              myPTRACE(1, "--> " << " " << cmd);
+
               HandleCmd(cmd, resp);
-              
-              if( resp.GetLength() ) {
+
+              if (resp.GetLength()) {
                 PBYTEArray _bresp((const BYTE *)(const char *)resp, resp.GetLength());
-              
+
                 myPTRACE(1, "<-- " << PRTHEX(_bresp));
                 bresp.Concatenate(_bresp);
               }
-        
+
               cmd = PString();
             }
           }
@@ -1431,7 +1655,9 @@ void ModemEngineBody::HandleData(const PBYTEArray &buf, PBYTEArray &bresp)
                 case -1:
                   state = stSendAckWait;
                   if( !t38engine || !t38engine->SendStop(moreFrames, NextSeq()) ) {
-                    bresp.Concatenate(PBYTEArray((const BYTE *)"\r\nERROR\r\n", 9));
+                    PString resp = RC_PREF() + RC_ERROR();
+
+                    bresp.Concatenate(PBYTEArray((const BYTE *)resp, resp.GetLength()));
                     state = stCommand;
                   }
                   break;
@@ -1473,13 +1699,18 @@ void ModemEngineBody::HandleData(const PBYTEArray &buf, PBYTEArray &bresp)
             PWaitAndSignal mutexWait(Mutex);
             state = stCommand;
             timeout.Stop();
+
+            PString resp = RC_PREF();
+
             if (t38engine) {
               t38engine->ResetModemState();
-              bresp.Concatenate(PBYTEArray((const BYTE *)"\r\nOK\r\n", 6));
+              resp += RC_OK();
             } else {
               _ClearCall();
-              bresp.Concatenate(PBYTEArray((const BYTE *)"\r\nNO CARRIER\r\n", 14));
+              resp += RC_NO_CARRIER();
             }
+
+            bresp.Concatenate(PBYTEArray((const BYTE *)resp, resp.GetLength()));
           }
       }
     }
@@ -1492,14 +1723,14 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
   {
     PWaitAndSignal mutexWait(Mutex);
     if (cmd.IsEmpty() && timerRing.Get())  {
-      resp += "\r\nRING\r\n";
+      resp = RC_RING();
       BYTE s0, ringCount;
       P.GetReg(0, s0);
       P.GetReg(1, ringCount);
       if (!ringCount && P.CidMode() == 10) {
-        resp += "NMBR = " + SrcNum();
-        resp += "\r\nNDID = " + DstNum();
-        resp += "\r\nRING\r\n";
+        resp += "NMBR = " + SrcNum() + "\r\n"
+                "NDID = " + DstNum() + "\r\n";
+        resp += RC_RING();
       }
       P.SetReg(1, ++ringCount);
       if (s0 > 0 && (ringCount >= s0)) {
@@ -1515,7 +1746,7 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
       case stConnectWait:
       case stReqModeAckWait:
       case stRecvBegWait:
-        resp += "\r\nNO CARRIER\r\n";
+        resp = RC_NO_CARRIER();
         state = stCommand;
         if (t38engine)
           t38engine->ResetModemState();
@@ -1536,22 +1767,22 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
           case stCommand:
             break;
           case stSend:
-            resp += "\r\nERROR\r\n";
+            resp = RC_ERROR();
             break;
           case stRecvBegWait:
-            resp += "\r\nNO CARRIER\r\n";
+            resp = RC_NO_CARRIER();
             break;
           case stConnectWait:
           case stConnectHandle:
           case stReqModeAckWait:
           case stReqModeAckHandle:
             if (callDirection == cdOutgoing)
-              resp += "\r\nBUSY\r\n";
+              resp = RC_BUSY();
             else
-              resp += "\r\nERROR\r\n";
+              resp = RC_ERROR();
             break;
           default:
-            resp += "\r\nERROR\r\n";
+            resp = RC_ERROR();
         }
         state = stCommand;
         if (t38engine)
@@ -1565,7 +1796,7 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
         PWaitAndSignal mutexWait(Mutex);
         state = stSendAckWait;
         if( !t38engine || !t38engine->SendStop(moreFrames, NextSeq()) ) {
-          resp += "\r\nERROR\r\n";
+          resp = RC_ERROR();
           state = stCommand;
         }
       }
@@ -1609,7 +1840,7 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
               } else {
                 state = stCommand;
                 timeout.Stop();
-                resp += "\r\nNO CARRIER\r\n";
+                resp = RC_NO_CARRIER();
               }
             }
             break;
@@ -1626,11 +1857,11 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
             if (t38engine && t38engine->SendStart(dataType, 3000)) {
               state = stSendAckWait;
               if (!t38engine->SendStop(FALSE, NextSeq())) {
-                resp += "\r\nERROR\r\n";
+                resp = RC_ERROR();
                 state = stCommand;
               }
             } else {
-              resp += "\r\nNO CARRIER\r\n";
+              resp = RC_NO_CARRIER();
               state = stCommand;
             }
             break;
@@ -1638,11 +1869,11 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
             if (t38engine)
               t38engine->SendOnIdle(T38Engine::dtCng);
             if( !RecvStart(T38Engine::dtHdlc, 3) ) {
-              resp += "\r\nERROR\r\n";
+              resp = RC_ERROR();
             }
             break;
           default:
-            resp += "\r\nNO CARRIER\r\n";
+            resp = RC_NO_CARRIER();
             state = stCommand;
         }
       }
@@ -1651,10 +1882,10 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
       switch( dataType ) {
         case T38Engine::dtCed:
           if( !SendStart(T38Engine::dtHdlc, 3, resp) )
-            resp += "\r\nERROR\r\n";
+            resp = RC_ERROR();
           break;
         case T38Engine::dtSilence:
-            resp += "\r\nOK\r\n";
+            resp = RC_OK();
             state = stCommand;
           break;
         case T38Engine::dtHdlc:
@@ -1662,10 +1893,10 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
           {
             PWaitAndSignal mutexWait(Mutex);
             if( moreFrames ) {
-              resp += "\r\nCONNECT\r\n";
+              resp = RC_CONNECT();
               state = stSend;
             } else {
-              resp += "\r\nOK\r\n";
+              resp = RC_OK();
               state = stCommand;
             }
             ResetDleData();
@@ -1679,20 +1910,23 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
         
         if( t38engine && t38engine->RecvStart(NextSeq()) ) {
           if( (t38engine->RecvDiag() & T38Engine::diagDiffSig) == 0 ) {
-            resp += "\r\nCONNECT\r\n";
+            if (dataType != T38Engine::dtRaw)
+              resp = RC_CONNECT();
             dataCount = 0;
             state = stRecv;
             parent.SignalDataReady();	// try to Recv w/o delay
           } else {
             t38engine->RecvStop();
-            resp += "\r\n+FCERROR\r\n";
+            resp = RC_FCERROR();
             state = stCommand;
           }
         } else {
-          resp += "\r\nERROR\r\n";
+          resp = RC_ERROR();
           state = stCommand;
         }
         ResetDleData();
+        if (dataType == T38Engine::dtHdlc)
+          fcs = FCS();
       }
       break;
     case stRecv:
@@ -1706,20 +1940,43 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
             break;
           }
           int count = t38engine->Recv(Buf, 1024);
-          
-          switch( count ) {
+
+          if (!dataCount && count && dataType == T38Engine::dtRaw) {
+            // send CONNECT just before data for AT+FRM command
+
+            PString _resp = RC_PREF() + RC_CONNECT();
+
+            PBYTEArray _bresp((const BYTE *)(const char *)_resp, _resp.GetLength());
+
+            myPTRACE(1, "<-- " << PRTHEX(_bresp));
+            bresp.Concatenate(_bresp);
+          }
+
+          switch (count) {
             case -1:
-              dleData.SetDiag(t38engine->RecvDiag()).PutEof();
-              t38engine->RecvStop();
-              if( dataCount == 0 && dataType == T38Engine::dtHdlc ) {
-                dleData.GetDleData(Buf, 1024);	// discard <DLE><ETX>
+              {
+                int diag = t38engine->RecvDiag();
+
+                if (dataType == T38Engine::dtHdlc) {
+                  Buf[0] = BYTE(fcs >> 8);
+                  Buf[1] = BYTE(fcs);
+                  if (diag & T38Engine::diagBadFcs)
+                     Buf[0]++;
+                  dleData.PutData(Buf, 2);
+                }
+                dleData.SetDiag(diag).PutEof();
               }
+              t38engine->RecvStop();
+              if (dataCount == 0 && dataType == T38Engine::dtHdlc)
+                dleData.GetDleData(Buf, 1024);	// discard ...<DLE><ETX>
               break;
             case 0:
               break;
             default:
               dataCount += count;
               dleData.PutData(Buf, count);
+              if (dataType == T38Engine::dtHdlc)
+                fcs.build(Buf, count);
           }
           if( count <= 0 ) break;
         }
@@ -1734,22 +1991,30 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
                 state = stCommand;
                 int diag = dleData.GetDiag();
                 
-                if( dataType == T38Engine::dtHdlc ) {
-                  if( diag == 0 )
-                    resp += "\r\nOK\r\n";
-                  else if( dataCount == 0 && (diag & ~T38Engine::diagNoCarrier) == 0 )
-                    resp += "\r\nNO CARRIER\r\n";
+                if (dataType == T38Engine::dtHdlc) {
+                  if (diag == 0)
+                    resp = RC_OK();
+                  else if (dataCount == 0 && (diag & ~T38Engine::diagNoCarrier) == 0)
+                    resp = RC_NO_CARRIER();
                   else
-                    resp += "\r\nERROR\r\n";
+                    resp = RC_ERROR();
                 } else {
-                  resp += "\r\nNO CARRIER\r\n";
+                  resp = RC_NO_CARRIER();
                 }
               }
               break;
             case 0:
               break;
             default:
-              PTRACE(1, "T38DLE--> " << PRTHEX(PBYTEArray(Buf, count)));
+#if PTRACING
+              if (myCanTrace(2)) {
+                 if (count <= 16) {
+                   PTRACE(2, "T38DLE--> " << PRTHEX(PBYTEArray(Buf, count)));
+                 } else {
+                   PTRACE(2, "T38DLE--> " << count << " bytes");
+                 }
+              }
+#endif
               bresp.Concatenate(PBYTEArray(Buf, count));
           }
           if( count <= 0 ) break;
@@ -1759,6 +2024,8 @@ void ModemEngineBody::CheckState(PBYTEArray & bresp)
   }
   
   if( resp.GetLength() ) {
+    resp = RC_PREF() + resp;
+
     PBYTEArray _bresp((const BYTE *)(const char *)resp, resp.GetLength());
     
     myPTRACE(1, "<-- " << PRTHEX(_bresp));
