@@ -3,7 +3,7 @@
  *
  * T38FAX Pseudo Modem
  *
- * Copyright (c) 2001-2002 Vyacheslav Frolov
+ * Copyright (c) 2001-2005 Vyacheslav Frolov
  *
  * Open H323 Project
  *
@@ -24,6 +24,32 @@
  * Contributor(s): Equivalence Pty ltd
  *
  * $Log: t38engine.cxx,v $
+ * Revision 1.34  2005/03/03 16:09:06  vfrolov
+ * Fixed memory leak
+ *
+ * Revision 1.33  2005/02/04 10:18:49  vfrolov
+ * Fixed warnings for No Trace build
+ *
+ * Revision 1.32  2005/02/03 11:32:12  vfrolov
+ * Fixed MSVC compile warnings
+ *
+ * Revision 1.31  2004/08/24 16:12:10  vfrolov
+ * Fixed bit counter overflow
+ *
+ * Revision 1.30  2004/07/06 16:07:24  vfrolov
+ * Included ptlib.h for precompiling
+ *
+ * Revision 1.29  2004/06/18 15:06:29  vfrolov
+ * Fixed race condition by adding mutex for modemCallback
+ *
+ * Revision 1.28  2004/03/01 17:10:32  vfrolov
+ * Fixed duplicated mutexes
+ * Added volatile to T38Mode
+ *
+ * Revision 1.27  2003/12/04 16:09:58  vfrolov
+ * Implemented FCS generation
+ * Implemented ECM support
+ *
  * Revision 1.26  2003/01/08 16:55:52  vfrolov
  * Added cbpOutBufNoFull and isOutBufFull()
  * Added data speed tracing
@@ -113,9 +139,10 @@
  *
  */
 
+#include <ptlib.h>
 #include "t38engine.h"
-#include "t38.h"
-#include "transports.h"
+#include <transports.h>
+#include <t38.h>
 
 #define new PNEW
 
@@ -134,18 +161,12 @@
 ///////////////////////////////////////////////////////////////
 enum StateOut {
   stOutIdle,
-  
   stOutCedWait,
   stOutSilenceWait,
   stOutIndWait,
-
   stOutData,
-  
   stOutHdlcFcs,
-  stOutHdlcFlagsWait,
-  
   stOutDataNoSig,
-
   stOutNoSig,
 };
 ///////////////////////////////////////////////////////////////
@@ -177,11 +198,13 @@ class ModStream
     BOOL SetDiag(int diag);
     BOOL PutEof(int diag = 0);
     void Move(ModStream &from);
-  
+
     DataStream *firstBuf;
     DataStreamQ bufQ;
     DataStream *lastBuf;	// if not NULL then shold be in bufQ or firstBuf
-    const MODPARS ModPars;
+    MODPARS ModPars;
+
+    HDLC hdlc;
 };
 
 ModStream::ModStream(const MODPARS &_ModPars) : firstBuf(NULL), lastBuf(NULL), ModPars(_ModPars)
@@ -190,12 +213,12 @@ ModStream::ModStream(const MODPARS &_ModPars) : firstBuf(NULL), lastBuf(NULL), M
 
 ModStream::~ModStream()
 {
-  if( firstBuf != NULL ) {
+  if (firstBuf != NULL) {
     PTRACE(1, "ModStream::~ModStream firstBuf != NULL, clean");
     delete firstBuf;
   }
-  if( bufQ.GetSize() > 0 )
-    PTRACE(1, "ModStream::~ModStream bufQ.GetSize()=" << bufQ.GetSize() << ", clean");
+  PTRACE_IF(1, bufQ.GetSize() > 0,
+    "ModStream::~ModStream bufQ.GetSize()=" << bufQ.GetSize() << ", clean");
 }
 
 void ModStream::PushBuf()
@@ -206,8 +229,8 @@ void ModStream::PushBuf()
 
 BOOL ModStream::DeleteFirstBuf()
 {
-  if( firstBuf != NULL ) {
-    if( lastBuf == firstBuf )
+  if (firstBuf != NULL) {
+    if (lastBuf == firstBuf)
       lastBuf = NULL;
     delete firstBuf;
     firstBuf = NULL;
@@ -218,17 +241,58 @@ BOOL ModStream::DeleteFirstBuf()
 
 BOOL ModStream::PopBuf()
 {
-  if( DeleteFirstBuf() )
+  if (DeleteFirstBuf()) {
     PTRACE(1, "ModStream::PopBuf DeleteFirstBuf(), clean");
-  return( (firstBuf = bufQ.Dequeue()) != NULL );
+  }
+  firstBuf = bufQ.Dequeue();
+  if (!firstBuf)
+    return FALSE;
+
+  if (ModPars.dataType == T38Engine::dtRaw && ModPars.dataTypeT38 == T38Engine::dtHdlc) {
+    hdlc = HDLC();
+    hdlc.PutHdlcData(firstBuf);
+    hdlc.GetRawStart(10);
+  }
+  return TRUE;
 }
 
 int ModStream::GetData(void *pBuf, PINDEX count)
 {
-  if( firstBuf == NULL ) {
+  if (firstBuf == NULL) {
     myPTRACE(1, "ModStream::GetData firstBuf == NULL");
     return -1;
   }
+
+  if (ModPars.dataType == T38Engine::dtRaw && ModPars.dataTypeT38 == T38Engine::dtHdlc) {
+    int len;
+
+    while ((len = hdlc.GetData(pBuf, count)) < 0) {
+      DataStream *_firstBuf;
+      if ((_firstBuf = bufQ.Dequeue()) != NULL) {
+        DeleteFirstBuf();
+        firstBuf = _firstBuf;
+        hdlc.PutHdlcData(firstBuf);
+        hdlc.GetRawStart();
+      } else {
+        if ((firstBuf->GetDiag() & T38Engine::diagNoCarrier) == 0) {
+          DeleteFirstBuf();
+          return 0;
+        }
+        return -1;
+      }
+    }
+    return len;
+  }
+
+  if (ModPars.dataTypeT38 != ModPars.dataType) {
+    myPTRACE(1, "ModStream::GetData ModPars.dataType("
+      << ModPars.dataType
+      << ") != ModPars.dataTypeT38("
+      << ModPars.dataTypeT38
+      << ")");
+    return -1;
+  }
+
   return firstBuf->GetData(pBuf, count);
 }
 
@@ -273,20 +337,27 @@ void ModStream::Move(ModStream &from)
   from.lastBuf = NULL;
 }
 ///////////////////////////////////////////////////////////////
+MODPARS::MODPARS(int _val, unsigned _ind, int _lenInd, unsigned _msgType, int _br)
+      : dataType(T38Engine::dtNone), dataTypeT38(T38Engine::dtNone),
+        val(_val), ind(_ind), lenInd(_lenInd),
+        msgType(_msgType), br(_br)
+{
+}
+
 static const MODPARS mods[] = {
-MODPARS( T38Engine::dtHdlc,  3, T38I(e_v21_preamble),              900, T38D(e_v21),         300 ),
-MODPARS( T38Engine::dtRaw,  24, T38I(e_v27_2400_training),        1100, T38D(e_v27_2400),   2400 ),
-MODPARS( T38Engine::dtRaw,  48, T38I(e_v27_4800_training),         900, T38D(e_v27_4800),   4800 ),
-MODPARS( T38Engine::dtRaw,  72, T38I(e_v29_7200_training),         300, T38D(e_v29_7200),   7200 ),
-MODPARS( T38Engine::dtRaw,  73, T38I(e_v17_7200_long_training),   1500, T38D(e_v17_7200),   7200 ),
-MODPARS( T38Engine::dtRaw,  74, T38I(e_v17_7200_short_training),   300, T38D(e_v17_7200),   7200 ),
-MODPARS( T38Engine::dtRaw,  96, T38I(e_v29_9600_training),         300, T38D(e_v29_9600),   9600 ),
-MODPARS( T38Engine::dtRaw,  97, T38I(e_v17_9600_long_training),   1500, T38D(e_v17_9600),   9600 ),
-MODPARS( T38Engine::dtRaw,  98, T38I(e_v17_9600_short_training),   300, T38D(e_v17_9600),   9600 ),
-MODPARS( T38Engine::dtRaw, 121, T38I(e_v17_12000_long_training),  1500, T38D(e_v17_12000), 12000 ),
-MODPARS( T38Engine::dtRaw, 122, T38I(e_v17_12000_short_training),  300, T38D(e_v17_12000), 12000 ),
-MODPARS( T38Engine::dtRaw, 145, T38I(e_v17_14400_long_training),  1500, T38D(e_v17_14400), 14400 ),
-MODPARS( T38Engine::dtRaw, 146, T38I(e_v17_14400_short_training),  300, T38D(e_v17_14400), 14400 ),
+MODPARS(   3, T38I(e_v21_preamble),              900, T38D(e_v21),         300 ),
+MODPARS(  24, T38I(e_v27_2400_training),        1100, T38D(e_v27_2400),   2400 ),
+MODPARS(  48, T38I(e_v27_4800_training),         900, T38D(e_v27_4800),   4800 ),
+MODPARS(  72, T38I(e_v29_7200_training),         300, T38D(e_v29_7200),   7200 ),
+MODPARS(  73, T38I(e_v17_7200_long_training),   1500, T38D(e_v17_7200),   7200 ),
+MODPARS(  74, T38I(e_v17_7200_short_training),   300, T38D(e_v17_7200),   7200 ),
+MODPARS(  96, T38I(e_v29_9600_training),         300, T38D(e_v29_9600),   9600 ),
+MODPARS(  97, T38I(e_v17_9600_long_training),   1500, T38D(e_v17_9600),   9600 ),
+MODPARS(  98, T38I(e_v17_9600_short_training),   300, T38D(e_v17_9600),   9600 ),
+MODPARS( 121, T38I(e_v17_12000_long_training),  1500, T38D(e_v17_12000), 12000 ),
+MODPARS( 122, T38I(e_v17_12000_short_training),  300, T38D(e_v17_12000), 12000 ),
+MODPARS( 145, T38I(e_v17_14400_long_training),  1500, T38D(e_v17_14400), 14400 ),
+MODPARS( 146, T38I(e_v17_14400_short_training),  300, T38D(e_v17_14400), 14400 ),
 };
 
 static const MODPARS invalidMods;
@@ -383,25 +454,29 @@ BOOL T38Engine::Attach(const PNotifier &callback)
 {
   PTRACE(1, name << " T38Engine::Attach");
   PWaitAndSignal mutexWaitModem(MutexModem);
+  PWaitAndSignal mutexWaitModemCallback(MutexModemCallback);
   PWaitAndSignal mutexWait(Mutex);
   if( !modemCallback.IsNULL() ) {
     myPTRACE(1, name << " T38Engine::Attach !modemCallback.IsNULL()");
     return FALSE;
   }
   modemCallback = callback;
-  ResetModemState();
+  _ResetModemState();
   return TRUE;
 }
 
 void T38Engine::Detach(const PNotifier &callback)
 {
-  PWaitAndSignal mutexWaitModem(MutexModem);
   PTRACE(1, name << " T38Engine::Detach");
+  PWaitAndSignal mutexWaitModem(MutexModem);
+  PWaitAndSignal mutexWaitModemCallback(MutexModemCallback);
   PWaitAndSignal mutexWait(Mutex);
   if( modemCallback == callback ) {
     modemCallback = NULL;
-    ResetModemState();
-    SetT38Mode(FALSE);
+    _ResetModemState();
+    T38Mode = FALSE;
+    myPTRACE(1, name << " T38Engine::Detach T38Mode=FALSE");
+    SignalOutDataReady();
   } else {
     myPTRACE(1, name << " T38Engine::Detach modemCallback != callback");
   }
@@ -438,7 +513,7 @@ void T38Engine::EncodeIFPPacket(PASN_OctetString &ifp_packet, const T38_IFPPacke
 BOOL T38Engine::Originate()
 {
   RenameCurrentThread(name + "(tx)");
-  PTRACE(3, "T38\tOriginate, transport=" << *transport);
+  PTRACE(2, "T38\tOriginate, transport=" << *transport);
 
   long seq = -1;
   int maxRedundancy = 0;
@@ -472,7 +547,7 @@ BOOL T38Engine::Originate()
       }
       else
         if (maxRedundancy > 0) {
-          PTRACE(4, "T38\tNot implemented yet " << recovery.GetTagName());
+          PTRACE(3, "T38\tNot implemented yet " << recovery.GetTagName());
         }
 
       udptl.m_seq_number = ++seq & 0xFFFF;
@@ -522,7 +597,7 @@ BOOL T38Engine::Originate()
         secondary.SetSize(nRedundancy);
       }
       else {
-        PTRACE(4, "T38\tNot implemented yet " << recovery.GetTagName());
+        PTRACE(3, "T38\tNot implemented yet " << recovery.GetTagName());
         continue;
       }
 #endif
@@ -542,8 +617,13 @@ BOOL T38Engine::Originate()
              << setprecision(2) << udptl << "\n  "
              << setprecision(2) << rawData);
       }
+      else
+      if (PTrace::CanTrace(3)) {
+        PTRACE(3, "T38\tSending PDU: seq=" << seq
+             << "\n  ifp = " << setprecision(2) << ifp);
+      }
       else {
-        PTRACE(3, "T38\tSending PDU:"
+        PTRACE(2, "T38\tSending PDU:"
                 " seq=" << seq <<
                 " type=" << ifp.m_type_of_msg.GetTagName());
       }
@@ -555,7 +635,7 @@ BOOL T38Engine::Originate()
              << setprecision(2) << rawData);
       }
       else {
-        PTRACE(3, "T38\tSending PDU again:"
+        PTRACE(2, "T38\tSending PDU again:"
                 " seq=" << seq);
       }
     }
@@ -577,7 +657,7 @@ BOOL T38Engine::Originate()
 BOOL T38Engine::Answer()
 {
   RenameCurrentThread(name + "(rx)");
-  PTRACE(3, "T38\tAnswer, transport=" << *transport);
+  PTRACE(2, "T38\tAnswer, transport=" << *transport);
 
   /* HACK HACK HACK -- need to figure out how to get the remote address
    * properly here */
@@ -631,13 +711,9 @@ BOOL T38Engine::Answer()
       receivedSequenceNumber -= 0x10000L;
     }
     
-#if PTRACING
-    if (PTrace::CanTrace(4)) {
-      PTRACE(4, "T38\tReceived PDU:\n  "
-             << setprecision(2) << rawData << "\n  UDPTL = "
-             << setprecision(2) << udptl);
-    }
-#endif
+    PTRACE(4, "T38\tReceived PDU:\n  "
+           << setprecision(2) << rawData << "\n  UDPTL = "
+           << setprecision(2) << udptl);
 
     if (lost < 0) {
       PTRACE(3, "T38\tRepeated packet");
@@ -661,7 +737,7 @@ BOOL T38Engine::Answer()
         receivedSequenceNumber -= lost;
 
         for (int i = nRedundancy - 1 ; i >= 0 ; i--) {
-          PTRACE(2, "T38\tReceived ifp seq=" << receivedSequenceNumber << " (secondary)");
+          PTRACE(3, "T38\tReceived ifp seq=" << receivedSequenceNumber << " (secondary)");
 
           if (!HandleRawIFP(secondary[i]))
             goto done;
@@ -673,7 +749,7 @@ BOOL T38Engine::Answer()
         receivedSequenceNumber += lost;
       }
       else {
-        PTRACE(4, "T38\tNot implemented yet " << recovery.GetTagName());
+        PTRACE(3, "T38\tNot implemented yet " << recovery.GetTagName());
       }
 
       if (lost) {
@@ -689,7 +765,7 @@ BOOL T38Engine::Answer()
     continue;
 #endif
 
-    PTRACE(4, "T38\tReceived ifp seq=" << receivedSequenceNumber);
+    PTRACE(3, "T38\tReceived ifp seq=" << receivedSequenceNumber);
 
     if (!HandleRawIFP(udptl.m_primary_ifp_packet))
       break;
@@ -710,38 +786,37 @@ done:
 //
 void T38Engine::ModemCallbackWithUnlock(INT extra)
 {
-  PNotifier callback = modemCallback;
-  
-  if (!callback.IsNULL()) {
-    Mutex.Signal();
-    callback(*this, extra);
-    Mutex.Wait();
+  Mutex.Signal();
+  {
+    PWaitAndSignal mutexWaitModemCallback(MutexModemCallback);
+    if (!modemCallback.IsNULL())
+      modemCallback(*this, extra);
   }
+  Mutex.Wait();
 }
 
 void T38Engine::CleanUpOnTermination()
 {
   myPTRACE(1, name << " T38Engine::CleanUpOnTermination");
   OpalT38Protocol::CleanUpOnTermination();
-  SetT38Mode(FALSE);
-}
 
-void T38Engine::SetT38Mode(BOOL mode)
-{
   PWaitAndSignal mutexWait(Mutex);
-  T38Mode = mode;
-  myPTRACE(1, name << " T38Engine::SetT38Mode T38Mode=" << (T38Mode ? "TRUE" : "FALSE"));
+  T38Mode = FALSE;
   SignalOutDataReady();
-  if (!T38Mode)
-    ModemCallbackWithUnlock(cbpReset);
+  ModemCallbackWithUnlock(cbpReset);
 }
 
 void T38Engine::ResetModemState() {
   PWaitAndSignal mutexWaitModem(MutexModem);
   PWaitAndSignal mutexWait(Mutex);
 
-  if (modStreamIn && modStreamIn->DeleteFirstBuf())
+  _ResetModemState();
+}
+
+void T38Engine::_ResetModemState() {
+  if (modStreamIn && modStreamIn->DeleteFirstBuf()) {
     PTRACE(1, name << " T38Engine::ResetModemState modStreamIn->DeleteFirstBuf(), clean");
+  }
 
   bufOut.PutEof();
   if (stateModem != stmIdle) {
@@ -794,24 +869,28 @@ BOOL T38Engine::SendStart(int _dataType, int param)
 
   switch( _dataType ) {
     case dtCed:
-      ModParsOut.dataType = _dataType;
+      ModParsOut.dataTypeT38 = ModParsOut.dataType = _dataType;
       ModParsOut.ind = T38I(e_ced);
       ModParsOut.lenInd = param;
       break;
     case dtSilence:
-      ModParsOut.dataType = _dataType;
+      ModParsOut.dataTypeT38 = ModParsOut.dataType = _dataType;
       ModParsOut.ind = T38I(e_no_signal);
       ModParsOut.lenInd = param;
       break;
     case dtHdlc:
     case dtRaw:
       ModParsOut = GetModPars(param);
-      if( !ModParsOut.IsModValid() || ModParsOut.dataType != _dataType ) 
+      ModParsOut.dataType = _dataType;
+      ModParsOut.dataTypeT38 =
+          (ModParsOut.msgType == T38D(e_v21) || t30.hdlcOnly()) ? dtHdlc : dtRaw;
+      if (!ModParsOut.IsModValid())
         return FALSE;
       break;
     default:
       return FALSE;
   }
+
   bufOut.Clean();		// reset eof
   stateModem = stmOutMoreData;
   SignalOutDataReady();
@@ -831,8 +910,9 @@ int T38Engine::Send(const void *pBuf, PINDEX count)
 
   PWaitAndSignal mutexWait(Mutex);
   int res = bufOut.PutData(pBuf, count);
-  if (res < 0)
+  if (res < 0) {
     myPTRACE(1, name << " T38Engine::Send res(" << res << ") < 0");
+  }
 
   SignalOutDataReady();
   return res;
@@ -854,6 +934,10 @@ BOOL T38Engine::SendStop(BOOL moreFrames, int _callbackParam)
   stateModem = stmOutNoMoreData;
   moreFramesOut = moreFrames;
   callbackParamOut = _callbackParam;
+
+  PTRACE(3, name << " T38Engine::SendStop moreFramesOut=" << moreFramesOut 
+                 << " callbackParamOut=" << callbackParamOut);
+
   SignalOutDataReady();
   return TRUE;
 }
@@ -868,55 +952,69 @@ BOOL T38Engine::RecvWait(int _dataType, int param, int _callbackParam)
     myPTRACE(1, name << " T38Engine::RecvWait stateModem(" << stateModem << ") != stmIdle");
     return FALSE;
   }
-  
-  MODPARS ModParsIn;
 
   PWaitAndSignal mutexWait(Mutex);
   switch( _dataType ) {
     case dtHdlc:
     case dtRaw:
-      ModParsIn = GetModPars(param);
-      if( !ModParsIn.IsModValid() || ModParsIn.dataType != _dataType )
-        return FALSE;
       break;
     default:
       return FALSE;
   }
 
+  const MODPARS &ModParsIn = GetModPars(param);
+
+  if (!ModParsIn.IsModValid())
+    return FALSE;
+
   callbackParamIn = _callbackParam;
 
-  if( modStreamIn != NULL ) {
-    if( modStreamIn->DeleteFirstBuf() )
+  if (modStreamIn != NULL) {
+    if (modStreamIn->DeleteFirstBuf()) {
       PTRACE(1, name << " T38Engine::RecvWait modStreamIn->DeleteFirstBuf(), clean");
+    }
   
-    if( modStreamIn->bufQ.GetSize() > 0 ) {
+    if (modStreamIn->bufQ.GetSize() > 0) {
       PTRACE(1, name << " T38Engine::RecvWait modStreamIn->bufQ.GetSize()=" << modStreamIn->bufQ.GetSize());
-      if( ModParsIn.val == modStreamIn->ModPars.val ) {
-        PTRACE(1, name << " T38Engine::RecvWait ModParsIn.val == modStreamIn->ModPars.val(" <<  modStreamIn->ModPars.val << ")");
+      if (ModParsIn.IsEqual(modStreamIn->ModPars)) {
+        PTRACE(1, name << " T38Engine::RecvWait ModParsIn == modStreamIn->ModPars("
+          << modStreamIn->ModPars.val
+          << ")");
+        modStreamIn->ModPars.dataType = _dataType;
+        modStreamIn->ModPars.dataTypeT38 =
+            (modStreamIn->ModPars.msgType == T38D(e_v21) || t30.hdlcOnly()) ? dtHdlc : dtRaw;
         stateModem = stmInReadyData;
         ModemCallbackWithUnlock(_callbackParam);
         return TRUE;
       }
-      delete modStreamIn;
-      modStreamIn = NULL;
     }
+    delete modStreamIn;
   }
 
   modStreamIn = new ModStream(ModParsIn);
-  
-  if( modStreamInSaved != NULL ) {
-    if( modStreamIn->ModPars.val == modStreamInSaved->ModPars.val ) {
-      myPTRACE(2, name << " T38Engine::RecvWait modStreamIn->ModPars.val(" <<  modStreamIn->ModPars.val << ") == modStreamInSaved->ModPars.val");
+  modStreamIn->ModPars.dataType = _dataType;
+  modStreamIn->ModPars.dataTypeT38 =
+      (modStreamIn->ModPars.msgType == T38D(e_v21) || t30.hdlcOnly()) ? dtHdlc : dtRaw;
+
+  if (modStreamInSaved != NULL) {
+    if (modStreamIn->ModPars.IsEqual(modStreamInSaved->ModPars)) {
+      myPTRACE(2, name << " T38Engine::RecvWait modStreamIn->ModPars == modStreamInSaved->ModPars("
+        << modStreamInSaved->ModPars.val
+        << ")");
       modStreamIn->Move(*modStreamInSaved);
       delete modStreamInSaved;
       modStreamInSaved = NULL;
     } else {
-      myPTRACE(2, name << " T38Engine::RecvWait modStreamIn->ModPars.val(" <<  modStreamIn->ModPars.val << ") != modStreamInSaved->ModPars.val(" << modStreamInSaved->ModPars.val << ")");
+      myPTRACE(2, name << " T38Engine::RecvWait modStreamIn->ModPars("
+        << modStreamIn->ModPars.val
+        << ") != modStreamInSaved->ModPars("
+        << modStreamInSaved->ModPars.val
+        << ")");
       modStreamIn->PushBuf();
       modStreamIn->PutEof(diagDiffSig);
     }
     stateModem = stmInReadyData;
-    ModemCallbackWithUnlock(callbackParamIn);
+    ModemCallbackWithUnlock(_callbackParam);
     return TRUE;
   }
 
@@ -937,8 +1035,10 @@ BOOL T38Engine::RecvStart(int _callbackParam)
   PWaitAndSignal mutexWait(Mutex);
   callbackParamIn = _callbackParam;
   
-  if( modStreamIn != NULL ) {
-    if( modStreamIn->PopBuf() ) {
+  if (modStreamIn != NULL) {
+    if (modStreamIn->PopBuf()) {
+      if (modStreamIn->ModPars.msgType == T38D(e_v21))
+        t30.v21Begin();
       stateModem = stmInRecvData;
       return TRUE;
     }
@@ -968,7 +1068,18 @@ int T38Engine::Recv(void *pBuf, PINDEX count)
     myPTRACE(1, name << " T38Engine::Recv modStreamIn == NULL");
     return -1;
   }
-  return modStreamIn->GetData(pBuf, count);
+
+  int len = modStreamIn->GetData(pBuf, count);
+
+  if (modStreamIn->ModPars.msgType == T38D(e_v21)) {
+    if (len > 0)
+      t30.v21Data(pBuf, len);
+    else
+    if (len < 0)
+      t30.v21End(FALSE);
+  }
+
+  return len;
 }
 
 int T38Engine::RecvDiag()
@@ -1034,12 +1145,11 @@ int T38Engine::PreparePacket(T38_IFPPacket & ifp, BOOL enableTimeout)
         case stOutIndWait:		outDelay = ModParsOut.lenInd; break;
 
         case stOutData:
-          outDelay = (countOut*8*1000)/ModParsOut.br + msPerOut - (PTime() - timeBeginOut).GetMilliSeconds();
+        case stOutHdlcFcs:
+          outDelay = int((PInt64(hdlcOut.getRawCount()) * 8 * 1000)/ModParsOut.br +
+                     msPerOut - (PTime() - timeBeginOut).GetMilliSeconds());
           break;
 
-        case stOutHdlcFcs:		outDelay = msPerOut; break;
-        case stOutHdlcFlagsWait:	outDelay = msPerOut*3; break;
-
         case stOutDataNoSig:		outDelay = msPerOut; break;
 
         case stOutNoSig:		outDelay = msPerOut; break;
@@ -1111,7 +1221,7 @@ int T38Engine::PreparePacket(T38_IFPPacket & ifp, BOOL enableTimeout)
 
               wasCarrierIn = FALSE;
 
-              switch( ModParsOut.dataType ) {
+              switch (ModParsOut.dataTypeT38) {
                 case dtHdlc:
                 case dtRaw:
                   t38indicator(ifp, ModParsOut.ind);
@@ -1126,7 +1236,7 @@ int T38Engine::PreparePacket(T38_IFPPacket & ifp, BOOL enableTimeout)
                   redo = TRUE;
                   break;
                 default:
-                  myPTRACE(1, "T38Engine::PreparePacket bad dataType=" << ModParsOut.dataType);
+                  myPTRACE(1, "T38Engine::PreparePacket bad dataTypeT38=" << ModParsOut.dataTypeT38);
                   return 0;
               }
               break;
@@ -1148,9 +1258,40 @@ int T38Engine::PreparePacket(T38_IFPPacket & ifp, BOOL enableTimeout)
             ////////////////////////////////////////////////////
             case stOutIndWait:
               stateOut = stOutData;
-              lastDteCharOut = -1;
               countOut = 0;
               timeBeginOut = PTime();
+              hdlcOut = HDLC();
+              if (ModParsOut.msgType == T38D(e_v21))
+                t30.v21Begin();
+
+              switch (ModParsOut.dataType) {
+                case dtHdlc:
+                  hdlcOut.PutHdlcData(&bufOut);
+                  break;
+                case dtRaw:
+                  hdlcOut.PutRawData(&bufOut);
+                  break;
+                default:
+                  myPTRACE(1, "T38Engine::PreparePacket bad dataType=" << ModParsOut.dataType);
+                  return 0;
+              }
+
+              switch (ModParsOut.dataTypeT38) {
+                case dtHdlc:
+                  hdlcOut.GetHdlcStart(TRUE);
+                  break;
+                case dtRaw:
+                  if (ModParsOut.dataType == dtHdlc) {
+                    myPTRACE(1, "T38Engine::PreparePacket sending dtHdlc like dtRaw not implemented");
+                    return 0;
+                  }
+                  hdlcOut.GetRawStart();
+                  break;
+                default:
+                  myPTRACE(1, "T38Engine::PreparePacket bad dataTypeT38=" << ModParsOut.dataTypeT38);
+                  return 0;
+              }
+
               redo = TRUE;
               break;
             ////////////////////////////////////////////////////
@@ -1161,84 +1302,99 @@ int T38Engine::PreparePacket(T38_IFPPacket & ifp, BOOL enableTimeout)
                 if (len > PINDEX(sizeof(b)))
                   len = sizeof(b);
                 BOOL wasFull = bufOut.isFull();
-                int count = bufOut.GetData(b, len);
+                int count = hdlcOut.GetData(b, len);
                 if (wasFull && !bufOut.isFull())
                   ModemCallbackWithUnlock(cbpOutBufNoFull);
 
                 switch( count ) {
                   case -1:
-                    switch( ModParsOut.dataType ) {
+                    switch (ModParsOut.dataTypeT38) {
                       case dtHdlc:
                         stateOut = stOutHdlcFcs;
                         break;
                       case dtRaw:
-#if PTRACING
-                        if (PTrace::CanTrace(2)) {
-                          long msTime = (PTime() - timeBeginOut).GetMilliSeconds();
-                          myPTRACE(2, "Sent " << countOut << " bytes in " << msTime << " ms (" << (countOut * 8 * 1000)/(msTime ? msTime : 1) << " bits/s)");
-                        }
-#endif
                         stateOut = stOutDataNoSig;
                         break;
                       default:
-                        myPTRACE(1, "T38Engine::PreparePacket stOutData bad dataType=" << ModParsOut.dataType);
+                        myPTRACE(1, "T38Engine::PreparePacket stOutData bad dataTypeT38=" << ModParsOut.dataTypeT38);
                         return 0;
                     }
                     redo = TRUE;
                     break;
                   case 0:
-                    if (lastDteCharOut != -1)
-                      if (ModParsOut.dataType == dtHdlc || lastDteCharOut != 0)
+                    if (hdlcOut.getLastChar() != -1)
+                      if (ModParsOut.dataType == dtHdlc || hdlcOut.getLastChar() != 0)
                         ModemCallbackWithUnlock(cbpOutBufEmpty);
                     waitData = TRUE;
                     break;
                   default:
-                    switch( ModParsOut.dataType ) {
+                    switch (ModParsOut.dataTypeT38) {
                       case dtHdlc:
+                        if (ModParsOut.msgType == T38D(e_v21)) {
+                          t30.v21Data(b, count);
+                        }
                         t38data(ifp, ModParsOut.msgType, T38F(e_hdlc_data), PBYTEArray(b, count));
                         break;
                       case dtRaw:
                         t38data(ifp, ModParsOut.msgType, T38F(e_t4_non_ecm_data), PBYTEArray(b, count));
                         break;
                       default:
-                        myPTRACE(1, "T38Engine::PreparePacket stOutData bad dataType=" << ModParsOut.dataType);
+                        myPTRACE(1, "T38Engine::PreparePacket stOutData bad dataTypeT38=" << ModParsOut.dataTypeT38);
                         return 0;
                     }
-                    lastDteCharOut = b[count - 1] & 0xFF;
                     countOut += count;
                 }
               }
               break;
             ////////////////////////////////////////////////////
             case stOutHdlcFcs:
-              if( stateModem != stmOutNoMoreData ) {
-                myPTRACE(1, "T38Engine::PreparePacket stOutHdlcFcs stateModem(" << stateModem << ") != stmOutNoMoreData");
-                return 0;
+              if (ModParsOut.msgType == T38D(e_v21)) {
+                t30.v21End(TRUE);
+                t30.v21Begin();
               }
-              t38data(ifp, ModParsOut.msgType, T38F(e_hdlc_fcs_OK));
-              if( moreFramesOut ) {
-                stateOut = stOutHdlcFlagsWait;
-                bufOut.Clean();		// reset eof
-                stateModem = stmOutMoreData;
-                ModemCallbackWithUnlock(callbackParamOut);
+              if (ModParsOut.dataType == dtRaw) {
+                if (countOut)
+                  t38data(ifp, ModParsOut.msgType, hdlcOut.isFcsOK() ? T38F(e_hdlc_fcs_OK) : T38F(e_hdlc_fcs_BAD));
+                hdlcOut.GetHdlcStart(FALSE);
+                countOut = 0;
+                if (hdlcOut.GetData(NULL, 0) != -1) {
+                  stateOut = stOutData;
+                } else {
+                  stateOut = stOutDataNoSig;
+                }
               } else {
-                stateOut = stOutDataNoSig;
+                if( stateModem != stmOutNoMoreData ) {
+                  myPTRACE(1, "T38Engine::PreparePacket stOutHdlcFcs stateModem(" << stateModem << ") != stmOutNoMoreData");
+                  return 0;
+                }
+                if (countOut)
+                  t38data(ifp, ModParsOut.msgType, T38F(e_hdlc_fcs_OK));
+                bufOut.Clean();		// reset eof
+                hdlcOut.PutHdlcData(&bufOut);
+                hdlcOut.GetHdlcStart(FALSE);
+                if (moreFramesOut) {
+                  stateOut = stOutData;
+                  stateModem = stmOutMoreData;
+                  ModemCallbackWithUnlock(callbackParamOut);
+                } else {
+                  stateOut = stOutDataNoSig;
+                }
               }
               break;
-            case stOutHdlcFlagsWait:
-              stateOut = stOutData;
-              lastDteCharOut = -1;
-              countOut = 0;
-              timeBeginOut = PTime();
-              redo = TRUE;
-              break;
             ////////////////////////////////////////////////////
             case stOutDataNoSig:
+#if PTRACING
+              if (myCanTrace(3) || (myCanTrace(2) && ModParsOut.dataType == dtRaw)) {
+                PInt64 msTime = (PTime() - timeBeginOut).GetMilliSeconds();
+                myPTRACE(2, "Sent " << hdlcOut.getRawCount() << " bytes in " << msTime << " ms ("
+                  << (PInt64(hdlcOut.getRawCount()) * 8 * 1000)/(msTime ? msTime : 1) << " bits/s)");
+              }
+#endif
               if( stateModem != stmOutNoMoreData ) {
                 myPTRACE(1, "T38Engine::PreparePacket stOutDataNoSig stateModem(" << stateModem << ") != stmOutNoMoreData");
                 return 0;
               }
-              switch( ModParsOut.dataType ) {
+              switch (ModParsOut.dataTypeT38) {
                 case dtHdlc:
                   t38data(ifp, ModParsOut.msgType, T38F(e_hdlc_sig_end));
                   break;
@@ -1246,7 +1402,7 @@ int T38Engine::PreparePacket(T38_IFPPacket & ifp, BOOL enableTimeout)
                   t38data(ifp, ModParsOut.msgType, T38F(e_t4_non_ecm_sig_end));
                   break;
                 default:
-                  myPTRACE(1, "T38Engine::PreparePacket stOutDataNoSig bad dataType=" << ModParsOut.dataType);
+                  myPTRACE(1, "T38Engine::PreparePacket stOutDataNoSig bad dataTypeT38=" << ModParsOut.dataTypeT38);
                   return 0;
               }
               stateOut = stOutNoSig;
@@ -1289,8 +1445,8 @@ int T38Engine::PreparePacket(T38_IFPPacket & ifp, BOOL enableTimeout)
       {
         PWaitAndSignal mutexWait(Mutex);
         if( stateOut == stOutData ) {
-          myPTRACE(1, "T38Engine::PreparePacket DTE's data delay, reset " << countOut);
-          countOut = 0;
+          myPTRACE(1, "T38Engine::PreparePacket DTE's data delay, reset " << hdlcOut.getRawCount());
+          hdlcOut.resetRawCount();
           timeBeginOut = PTime() - PTimeInterval(msPerOut);
           doDalay = FALSE;
         }
@@ -1301,7 +1457,7 @@ int T38Engine::PreparePacket(T38_IFPPacket & ifp, BOOL enableTimeout)
   return 1;
 }
 ///////////////////////////////////////////////////////////////
-BOOL T38Engine::HandlePacketLost(unsigned nLost)
+BOOL T38Engine::HandlePacketLost(unsigned myPTRACE_PARAM(nLost))
 {
   PWaitAndSignal mutexWaitIn(MutexIn);
 
@@ -1327,12 +1483,12 @@ BOOL T38Engine::HandlePacketLost(unsigned nLost)
 BOOL T38Engine::HandlePacket(const T38_IFPPacket & ifp)
 {
 #if PTRACING
-  if (PTrace::CanTrace(4)) {
-    PTRACE(4, "T38\tReceived ifp\n  "
+  if (PTrace::CanTrace(3)) {
+    PTRACE(3, "T38\tReceived ifp\n  "
              << setprecision(2) << ifp);
   }
   else {
-    PTRACE(3, "T38\tReceived ifp type=" << ifp.m_type_of_msg.GetTagName());
+    PTRACE(2, "T38\tReceived ifp type=" << ifp.m_type_of_msg.GetTagName());
   }
 #endif
 
@@ -1381,14 +1537,16 @@ BOOL T38Engine::HandlePacket(const T38_IFPPacket & ifp)
 
           if( stateModem == stmInWaitData ) {
             if( modStreamIn != NULL ) {
-              if( modStreamIn->ModPars.val == modStreamInSaved->ModPars.val ) {
+              if (modStreamIn->ModPars.IsEqual(modStreamInSaved->ModPars)) {
                 modStreamIn->Move(*modStreamInSaved);
                 delete modStreamInSaved;
                 modStreamInSaved = NULL;
               } else {
-                myPTRACE(2, "T38Engine::HandlePacket modStreamIn->ModPars.val("
-                  << modStreamIn->ModPars.val << ") != modStreamInSaved->ModPars.val("
-                  << modStreamInSaved->ModPars.val << ")");
+                myPTRACE(2, name << " T38Engine::HandlePacket modStreamIn->ModPars("
+                  << modStreamIn->ModPars.val
+                  << ") != modStreamInSaved->ModPars("
+                  << modStreamInSaved->ModPars.val
+                  << ")");
                 modStreamIn->PushBuf();
                 modStreamIn->PutEof(diagDiffSig);
               }
@@ -1441,7 +1599,7 @@ BOOL T38Engine::HandlePacket(const T38_IFPPacket & ifp)
                   break;
                 }
                 const T38_Data_Field_subtype &Data_Field = ifp.m_data_field[i];
-                
+
                 switch( Data_Field.m_field_type ) {	// Handle data
                   case T38F(e_hdlc_data):
                   case T38F(e_t4_non_ecm_data):
@@ -1464,18 +1622,13 @@ BOOL T38Engine::HandlePacket(const T38_IFPPacket & ifp)
                   default:
                     myPTRACE(1, "T38Engine::HandlePacket field_type bad !!! " << setprecision(2) << ifp);
                 }
-                switch( Data_Field.m_field_type ) {	// Handle fcs_BAD
+                switch( Data_Field.m_field_type ) {	// Handle fcs
                   case T38F(e_hdlc_fcs_BAD):
                   case T38F(e_hdlc_fcs_BAD_sig_end):
                     modStream->SetDiag(diagBadFcs);
                     myPTRACE(1, "T38Engine::HandlePacket bad FCS");
-                    break;
-                }
-                switch( Data_Field.m_field_type ) {	// Handle fcs
                   case T38F(e_hdlc_fcs_OK):
-                  case T38F(e_hdlc_fcs_BAD):
                   case T38F(e_hdlc_fcs_OK_sig_end):
-                  case T38F(e_hdlc_fcs_BAD_sig_end):
                     modStream->PutEof();
                     modStream->PushBuf();
                     break;
@@ -1483,9 +1636,10 @@ BOOL T38Engine::HandlePacket(const T38_IFPPacket & ifp)
                 switch( Data_Field.m_field_type ) {	// Handle sig_end
                   case T38F(e_t4_non_ecm_sig_end):
 #if PTRACING
-                    if (PTrace::CanTrace(2)) {
-                      long msTime = (PTime() - timeBeginIn).GetMilliSeconds();
-                      myPTRACE(2, "Received " << countIn << " bytes in " << msTime << " ms (" << (countIn * 8 * 1000)/(msTime ? msTime : 1) << " bits/s)");
+                    if (myCanTrace(2)) {
+                      PInt64 msTime = (PTime() - timeBeginIn).GetMilliSeconds();
+                      myPTRACE(2, "Received " << countIn << " bytes in " << msTime << " ms ("
+                        << (PInt64(countIn) * 8 * 1000)/(msTime ? msTime : 1) << " bits/s)");
                     }
 #endif
                   case T38F(e_hdlc_fcs_OK_sig_end):
